#!/usr/bin/python

import os
import sys
import time
import yaml
import socket

from contextlib import contextmanager, closing
from subprocess import check_output


class VaultUnexpectedlySealedError(Exception):
    pass


class VaultUnexpectedlyUnSealedError(Exception):
    pass


class VaultReadYieldedUnexpectedValue(Exception):
    pass


vault_init_cmd = 'sh -c "if [ -s vault.init.out ]; then ' \
                 'cat vault.init.out; ' \
                 'else env VAULT_ADDR=https://127.0.0.1:8200/ VAULT_SKIP_VERIFY=1 ' \
                 '/snap/bin/vault init ' \
                 '| tee vault.init.out ; fi"'

vault_cmd = 'env VAULT_ADDR=https://127.0.0.1:8200/ VAULT_SKIP_VERIFY=1 /snap/bin/vault'

vault_auth_cmd_template = vault_cmd + ' auth {token}'
vault_read_cmd_template = vault_cmd + ' read -field={key} {path}'
vault_status_cmd = vault_cmd + ' status || true'  # because rc=2 when sealed
vault_unseal_cmd_template = vault_cmd + ' unseal {unseal_key}'
vault_write_cmd_template = vault_cmd + ' write {path} {key}={value}'


def juju_run(unit, command):
    return ['juju', 'run', '--unit', unit, command]


def get_fields(output, field_prefix):
    return [line.split(': ')[1]
            for line in output.split('\n')
            if line.startswith(field_prefix)]


def get_unseal_keys(init_output):
    return get_fields(init_output, 'Unseal Key')


def get_root_token(init_output):
    return get_fields(init_output, 'Initial Root Token')[0]


@contextmanager
def doing(message, where=sys.stdout):
    where.write(message + '... ')
    where.flush()
    yield
    where.write('done.\n')
    where.flush()


def main():
    do_init()
    do_write_test()
    do_read_test()
    do_simple_restore_test()


def do_init():
    with doing('Scanning Juju model for units, like a Cylon'):
        status = yaml.safe_load(check_output(
            ['juju', 'status', '--format=yaml']))
        vault_unit = list(status['applications']['vault']['units'].keys())[0]
        postgres_unit = list(status['applications']['postgresql']['units'].keys())[0]

    with doing('Initializing vault'):
        init_output = check_output(juju_run(vault_unit, vault_init_cmd))
    unseal_keys = get_unseal_keys(init_output)
    root_token = get_root_token(init_output)

    with doing('Storing "vault init" output in Mojo workspace'):
        with open(os.path.join(os.environ['MOJO_LOCAL_DIR'], 'vault.init.output'), 'w') as f:
            f.write(init_output)


def do_write_test():
    with doing('Scanning Juju model for units, like a Cylon'):
        status = yaml.safe_load(check_output(
            ['juju', 'status', '--format=yaml']))
        vault_unit = list(status['applications']['vault']['units'].keys())[0]
        postgres_unit = list(status['applications']['postgresql']['units'].keys())[0]

    with doing('Reading "vault init" output from Mojo workspace'):
        with open(os.path.join(os.environ['MOJO_LOCAL_DIR'], 'vault.init.output')) as f:
            init_output = f.read()
    unseal_keys = get_unseal_keys(init_output)
    root_token = get_root_token(init_output)

    with doing('Checking whether vault is sealed'):
        status_output = check_output(juju_run(vault_unit, vault_status_cmd))
        sealed = get_fields(status_output, 'Sealed')[0]

    if sealed == 'true':
        with doing('Unsealing vault'):
            for key in unseal_keys[:3]:
                check_output(juju_run(
                    vault_unit, vault_unseal_cmd_template.format(unseal_key=key)))

        with doing('Checking vault is unsealed'):
            status_output = check_output(juju_run(vault_unit, vault_status_cmd))
            sealed = get_fields(status_output, 'Sealed')[0]
            if sealed != 'false':
                raise VaultUnexpectedlySealedError(status_output)

    with doing('Authenticating to vault'):
        check_output(juju_run(
            vault_unit, vault_auth_cmd_template.format(token=root_token)))

    with doing('Writing to vault'):
        check_output(juju_run(vault_unit, vault_write_cmd_template.format(
            path='/secret', key='unit', value=vault_unit)))

    with doing('Restarting vault'):
        check_output(juju_run(vault_unit, 'systemctl restart vault'))

    # TODO(pjdc): Might end up needing a sleep or suchlike here.

    with doing('Checking whether vault is re-sealed'):
        status_output = check_output(juju_run(vault_unit, vault_status_cmd))
        sealed = get_fields(status_output, 'Sealed')[0]
        if sealed != 'true':
            raise VaultUnexpectedlyUnSealedError(status_output)


def do_read_test():
    with doing('Scanning Juju model for units, like a Cylon'):
        status = yaml.safe_load(check_output(
            ['juju', 'status', '--format=yaml']))
        vault_unit = list(status['applications']['vault']['units'].keys())[0]
        postgres_unit = list(status['applications']['postgresql']['units'].keys())[0]

    with doing('Reading "vault init" output from Mojo workspace'):
        with open(os.path.join(os.environ['MOJO_LOCAL_DIR'], 'vault.init.output')) as f:
            init_output = f.read()
    unseal_keys = get_unseal_keys(init_output)
    root_token = get_root_token(init_output)

    with doing('Unsealing vault'):
        for key in unseal_keys[:3]:
            check_output(juju_run(
                vault_unit, vault_unseal_cmd_template.format(unseal_key=key)))

    with doing('Checking vault is unsealed'):
        status_output = check_output(juju_run(vault_unit, vault_status_cmd))
        sealed = get_fields(status_output, 'Sealed')[0]
        if sealed != 'false':
            raise VaultUnexpectedlySealedError(status_output)

    with doing('Authenticating to vault'):
        check_output(juju_run(
            vault_unit, vault_auth_cmd_template.format(token=root_token)))

    with doing('Reading from vault'):
        value = check_output(juju_run(vault_unit, vault_read_cmd_template.format(
            path='/secret', key='unit')))
        if value != vault_unit:
            raise VaultReadYieldedUnexpectedValue(
                'retrieved value "{}" != expected value "{}"'.format(
                    value, vault_unit))


def do_simple_restore_test():
    with doing('Scanning Juju model for units, like a Cylon'):
        status = yaml.safe_load(check_output(
            ['juju', 'status', '--format=yaml']))
        vault_unit = list(status['applications']['vault']['units'].keys())[0]
        postgres_unit = list(status['applications']['postgresql']['units'].keys())[0]

    with doing('Reading "vault init" output from Mojo workspace'):
        with open(os.path.join(os.environ['MOJO_LOCAL_DIR'], 'vault.init.output')) as f:
            init_output = f.read()
    unseal_keys = get_unseal_keys(init_output)
    root_token = get_root_token(init_output)

    with doing('Backing up vault database'):
        check_output(juju_run(postgres_unit, 'sudo -u postgres pg_dump vault > /tmp/vault.dump'))

    with doing('Stopping vault'):
        check_output(juju_run(vault_unit, 'systemctl stop vault'))

    with doing('Dropping vault database'):
        check_output(juju_run(postgres_unit, 'sudo -u postgres dropdb vault'))

    with doing('Creating blank vault database'):
        check_output(juju_run(postgres_unit, 'sudo -u postgres createdb vault'))

    with doing('Restoring vault database'):
        check_output(juju_run(postgres_unit, 'sudo -u postgres psql -f /tmp/vault.dump vault'))

    with doing('Starting vault'):
        check_output(juju_run(vault_unit, 'systemctl start vault'))

    with doing('Checking whether vault is sealed'):
        status_output = check_output(juju_run(vault_unit, vault_status_cmd))
        sealed = get_fields(status_output, 'Sealed')[0]
        if sealed != 'true':
            raise VaultUnexpectedlyUnSealedError(status_output)

    with doing('Unsealing vault'):
        for key in unseal_keys[:3]:
            check_output(juju_run(
                vault_unit, vault_unseal_cmd_template.format(unseal_key=key)))

    with doing('Checking vault is unsealed'):
        status_output = check_output(juju_run(vault_unit, vault_status_cmd))
        sealed = get_fields(status_output, 'Sealed')[0]
        if sealed != 'false':
            raise VaultUnexpectedlySealedError(status_output)

    with doing('Authenticating to vault'):
        check_output(juju_run(
            vault_unit, vault_auth_cmd_template.format(token=root_token)))

    with doing('Reading from vault'):
        value = check_output(juju_run(vault_unit, vault_read_cmd_template.format(
            path='/secret', key='unit')))
        if value != vault_unit:
            raise VaultReadYieldedUnexpectedValue(
                'retrieved value "{}" != expected value "{}"'.format(
                    value, vault_unit))


def do_restart_unseal_test():
    with doing('Scanning Juju model for units, like a Cylon'):
        status = yaml.safe_load(check_output(
            ['juju', 'status', '--format=yaml']))
        vault_unit = list(status['applications']['vault']['units'].keys())[0]
        postgres_unit = list(status['applications']['postgresql']['units'].keys())[0]

    with doing('Reading "vault init" output from Mojo workspace'):
        with open(os.path.join(os.environ['MOJO_LOCAL_DIR'], 'vault.init.output')) as f:
            init_output = f.read()
    unseal_keys = get_unseal_keys(init_output)
    root_token = get_root_token(init_output)

    with doing('Ensuring we can restart vault'):
        check_output(juju_run(vault_unit, 'systemctl reset-failed vault'))

    with doing('Restarting vault'):
        check_output(juju_run(vault_unit, 'systemctl restart vault'))

    # We need to make sure we can connect before we try again.
    with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
        for t in range(15):
            with doing('Sleeping for {} seconds to let the restart complete'.format(t)):
                time.sleep(t)
            # connect_ex returns 0 on success
            if not s.connect_ex(('127.0.0.1', 8200)):
                break  # Vault is finally listening, so we can unseal it.

    with doing('Unsealing vault'):
        for key in unseal_keys[:3]:
            check_output(juju_run(
                vault_unit, vault_unseal_cmd_template.format(unseal_key=key)))

    with doing('Checking vault is unsealed'):
        status_output = check_output(juju_run(vault_unit, vault_status_cmd))
        sealed = get_fields(status_output, 'Sealed')[0]
        if sealed != 'false':
            raise VaultUnexpectedlySealedError(status_output)

    with doing('Authenticating to vault'):
        check_output(juju_run(
            vault_unit, vault_auth_cmd_template.format(token=root_token)))


actions = {
    'init': do_init,
    'write_test': do_write_test,
    'read_test': do_read_test,
    'simple_restore_test': do_simple_restore_test,
    'restart_unseal_test': do_restart_unseal_test,
}

if __name__ == '__main__':
    action = actions.get(os.environ.get('action'))
    if action:
        action()
    else:
        main()
