# Modules
import requests


class LocalEnvironmentSwiftContainer:
    """
    Manage a Swift container for the local deployment environment
    for storing deployment settings

    swift_connection: A working swiftclient.client.Connection object
    container_name: The name of the swift container to use
    """

    previous_build_obj = 'previous-build-label'
    deployed_build_obj = 'deployed-build-label'
    deployed_revno_obj = 'deployed-spec-revno'
    code_upgrade_succeeded_template = 'code-upgrade-{}-{}-succeeded'
    mojo_run_succeeded_template = 'mojo-run-{}-succeeded'

    def __init__(self, swift_connection, container_name):
        self.swift_connection = swift_connection
        self.container_name = container_name

    def previous_build_label(self):
        """
        Get the build label that was previously deployed on this environment
        From the Swift account associated with this environment
        through environment variables "OS_AUTH_URL" etc.
        """

        return self.swift_connection.get_object(
            container=self.container_name,
            obj=self.previous_build_obj
        )[1].strip()

    def deployed_build_label(self):
        """
        Get the build label that is currently deployed on this environment
        From the Swift account associated with this environment
        through environment variables "OS_AUTH_URL" etc.
        """

        return self.swift_connection.get_object(
            container=self.container_name,
            obj=self.deployed_build_obj
        )[1].strip()

    def deployed_spec_revno(self):
        """
        Get the mojo spec revision number that is currently deployed
        on this environment from the Swift account associated with this
        environment through environment variables "OS_AUTH_URL" etc.
        """

        return self.swift_connection.get_object(
            container=self.container_name,
            obj=self.deployed_revno_obj
        )[1].strip()

    def save_code_upgrade_succeeded(
        self, from_build_label, to_build_label
    ):
        """
        Save an object into Swift
        to signify that a code-upgrade succeeded
        from one build_label to another
        """

        upgrade_object_name = self.code_upgrade_succeeded_template.format(
            from_build_label, to_build_label
        )

        self.swift_connection.put_object(
            container=self.container_name,
            obj=upgrade_object_name,
            contents='true'
        )

    def save_mojo_run_succeeded(self, spec_revno):
        """
        Save an object into Swift
        to signifiy that a mojo-run succeeded with
        a given revision number of the mojo spec
        """

        run_object_name = self.mojo_run_succeeded_template.format(
            spec_revno
        )

        self.swift_connection.put_object(
            container=self.container_name,
            obj=run_object_name,
            contents='true'
        )

    def save_deployed_build_label(self, build_label):
        """
        Save an object into swift containing
        the build_label which is deployed to this environment
        """

        self.swift_connection.put_object(
            container=self.container_name,
            obj='deployed-build-label',
            contents=build_label
        )

    def save_previous_build_label(self, build_label):
        """
        Save an object into swift
        to record the previously deployed build_label
        """

        self.swift_connection.put_object(
            container=self.container_name,
            obj='previous-build-label',
            contents=build_label
        )

    def save_mojo_spec_revno(self, spec_revno):
        """
        Save an object into swift containing
        the current mojo spec revision number
        """

        self.swift_connection.put_object(
            container=self.container_name,
            obj='deployed-spec-revno',
            contents=spec_revno
        )


class HttpContainer:
    """
    Methods for retrieving objects
    from a Swift HTTP storage container
    """

    def __init__(self, container_url):
        """
        container_url: The storage URL path for the swift container
        """

        self.container_url = container_url

    def get(self, object_name):
        """
        Retrieve the contents of an object
        """

        object_url = '{}/{}'.format(self.container_url, object_name)

        response = requests.get(object_url)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            http_error.message += '. URL: {}'.format(object_url)
            http_error.args += ('URL: {}'.format(object_url),)
            raise http_error

        response.raise_for_status()

        return response.text

    def head(self, object_name):
        """
        Retrieve the HEAD of an object
        """

        object_url = '{}/{}'.format(self.container_url, object_name)

        return requests.head(object_url)

    def exists(self, object_name):
        """
        Check an object exists
        """

        return self.head(object_name).ok


class CIContainer(HttpContainer):
    """
    Methods for retrieving continuous integration
    resources from an http swift store
    """

    code_upgrade_succeeded_template = 'code-upgrade-{}-{}-succeeded'
    mojo_run_succeeded_template = 'mojo-run-{}-succeeded'

    def has_code_upgrade_been_tested(
        self,
        from_build_label,
        to_build_label
    ):
        """
        Check if a specific code upgrade has been tested
        (from one build_label to another)
        by checking if a specially named Swift object exists
        in the CI system's Swift HTTP store.
        (This object will have been created
        by self.save_code_upgrade_succeeded)
        """

        return self.exists(
            self.code_upgrade_succeeded_template.format(
                from_build_label,
                to_build_label
            )
        )

    def has_mojo_run_been_tested(self, spec_revno):
        """
        Check if a specific mojo spec revision number has been tested
        by checking if a specially named Swift object exists
        in the CI system's Swift HTTP store.
        (This object will have been created by self.save_mojo_run_succeeded)
        """

        return self.exists(
            self.mojo_run_succeeded_template.format(
                spec_revno
            )
        )


class BuildContainer(HttpContainer):
    """
    Methods for interacting with the HTTP Swift store
    containing code builds
    """

    def latest_build_label(self):
        """
        Get the latest build label from the webteam's Swift HTTP object store
        """

        return self.get('latest-build-label').strip()


class DeployedEnvironmentContainer(HttpContainer):
    """
    Methods for requesting information
    from the HTTP swift store for a
    depoyed environment (e.g. production, staging)
    """

    def deployed_build_label(self):
        """
        Get the build label which was most recently
        deployed to this environment
        from their Swift HTTP object store
        """

        return self.get('deployed-build-label').strip()

    def deployed_spec_revno(self):
        """
        Get the version of the mojo spec
        which was most recently run on this environment
        from their Swift HTTP object store
        """

        return self.get('deployed-spec-revno').strip()
