# System
import os

# Local imports
from container_managers import (
    BuildContainer,
    CIContainer,
    DeployedEnvironmentContainer,
    LocalEnvironmentSwiftContainer
)
from mojo_utils import build_swift_connection

"""
Tests for container_managers.

To test these you'll need to set a TEST_CONTAINER_URL,
which should be the HTTP URL to a swift container which contains
the following publicly readable objects:
- latest-build-label
- deployed-build-label
- deployed-spec-revno
- code-upgrade-test-build-label-01-test-build-label-02-succeeded
- mojo-run-155-succeeded

You'll also need to have the following environment variables setup
with credentials to connect to a valid (testing) Swift account:

OS_AUTH_URL, OS_USERNAME, OS_PASSWORD, OS_TENANT_NAME, OS_REGION_NAME
"""

test_container_url = os.environ.get('TEST_CONTAINER_URL')


def test_deployed_environment_container():
    print (
        "\n===\n"
        "Test DeployedEnvironmentContainer"
        "\n===\n"
    )

    deployed_env_container = DeployedEnvironmentContainer(
        container_url=test_container_url,
    )

    build_label = deployed_env_container.deployed_build_label()
    assert build_label, "Build label missing"
    print u"\u2713 Found build label: {}".format(build_label)

    spec_revno = deployed_env_container.deployed_spec_revno()
    assert spec_revno, "Spec revno missing"
    print u"\u2713 Found spec revno (from web): {}".format(spec_revno)


def test_local_environment_swift_container():
    print (
        "\n===\n"
        "Test LocalEnvironmentSwiftContainer"
        "\n===\n"
    )

    swift_connection = build_swift_connection()

    container_name = 'test-container'

    swift_connection.put_container(container_name)

    container = LocalEnvironmentSwiftContainer(
        swift_connection=swift_connection,
        container_name=container_name
    )

    deployed_build_label = "fiddlesticksandfishes"
    previous_build_label = "fishesandfiddlesticks"
    deployed_spec_revno = '111'

    container.save_deployed_build_label(deployed_build_label)
    container.save_previous_build_label(previous_build_label)
    container.save_mojo_spec_revno(deployed_spec_revno)

    assert container.deployed_build_label() == deployed_build_label
    assert container.previous_build_label() == previous_build_label
    assert container.deployed_spec_revno() == deployed_spec_revno
    print u"\u2713 Saved and retrieved build_labels and spec_revno"

    # Get names for objects
    previous_obj = container.previous_build_obj
    deployed_obj = container.deployed_build_obj
    revno_obj = container.deployed_revno_obj
    upgrade_object = container.code_upgrade_succeeded_template.format(
        previous_build_label, deployed_build_label
    )
    run_object = container.mojo_run_succeeded_template.format(
        deployed_spec_revno
    )

    # Remove the objects if they exist
    for object_name in [previous_obj, deployed_obj, revno_obj, upgrade_object]:
        try:
            swift_connection.delete_object(container_name, object_name)
        except:
            pass

    container.save_code_upgrade_succeeded(
        previous_build_label, deployed_build_label
    )

    container.save_mojo_run_succeeded(deployed_spec_revno)

    upgrade_head = swift_connection.head_object(container_name, upgrade_object)

    run_head = swift_connection.head_object(container_name, run_object)

    # Make sure the objects were saved to swift
    assert int(upgrade_head['content-length']) > 0
    assert int(run_head['content-length']) > 0

    print u"\u2713 Code upgrade and mojo run successfully saved"


def test_latest_build():
    print (
        "\n===\n"
        "Test BuildContainer"
        "\n===\n"
    )

    build_container = BuildContainer(
        container_url=test_container_url
    )

    build_label = build_container.latest_build_label()
    assert build_label, "Build label missing"
    print u"\u2713 Found latest build label: {}".format(build_label)


def test_ci_container():
    print (
        "\n===\n"
        "Test CIContainer"
        "\n===\n"
    )

    ci_container = CIContainer(
        container_url=test_container_url
    )

    build_one = "test-build-label-01"
    build_two = "test-build-label-02"

    good_upgrade = ci_container.has_code_upgrade_been_tested(
        from_build_label=build_one,
        to_build_label=build_two
    )
    bad_upgrade = ci_container.has_code_upgrade_been_tested(
        from_build_label=build_two,
        to_build_label=build_one
    )

    good_upgrade = ci_container.has_code_upgrade_been_tested(
        from_build_label=build_one,
        to_build_label=build_two
    )
    bad_upgrade = ci_container.has_code_upgrade_been_tested(
        from_build_label=build_two,
        to_build_label=build_one
    )

    assert good_upgrade, "Good upgrade test missing"
    assert bad_upgrade is False, "Bad upgrade isn't false"
    print u"\u2713 Upgrades checked successfully"

    good_mojo_test = ci_container.has_mojo_run_been_tested(spec_revno="155")
    bad_mojo_test = ci_container.has_mojo_run_been_tested(spec_revno="99999")

    assert good_mojo_test, "Good mojo test missing"
    assert bad_mojo_test is False, "Bad mojo test isn't false"
    print u"\u2713 Mojo runs checked succeeded"


def container_tests():
    test_deployed_environment_container()
    test_local_environment_swift_container()
    test_latest_build()
    test_ci_container()


if __name__ == "__main__":
    container_tests()
